/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Jetpack Toolbar code.
 *
 * The Initial Developer of the Original Code is Mozilla.
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   David Dahl <ddahl@mozilla.com> (Original Author)
 *   Paul O’Shannessy <paul@oshannessy.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// toolbar api

// CONSTANTS
const TOOLBAR_NORMAL = 0;
const TOOLBAR_BOOKMARKS = 1;

const ERRORS = {
  BUTTON_OPTIONS_ID_REQUIRED:
    "Invalid button options: An id is required.",
  BUTTON_OPTIONS_UNIQUE_ID:
    "Invalid button options: This id is already in use",
  BUTTON_OPTIONS_ONCLICK_REQUIRED:
    "Invalid button options: An onclick function is required for this type of toolbar button.",
  BUTTON_OPTIONS_URL_REQUIRED:
    "Invalid button options: A url is required for this type of toolbar button.",
  BUTTON_NOT_FOUND:
    "A button was not found with that id."
};

function Toolbar(featureContext) {
  this._featureContext = featureContext;
  this._unloaders = [];

  Extension.addUnloadMethod(this, function() {
    while (this._unloaders.length)
      this._unloaders.shift().call();
  });
}

Toolbar.prototype = {
  _featureContext: null,
  _unloaders: null,

  get navigation() {
    let bar = new AtulBar({
      id: "nav-bar",
      type: TOOLBAR_NORMAL,
      toolbar: this
    });
    let _navigation = {
      append: function(options) {
        bar.append(options);
      },
      remove: function(buttonID) {
        bar.remove(buttonID);
      }
    };
    this._unloaders.push(function() bar.unload());

    this.__defineGetter__("navigation", function() _navigation);
    return this.navigation;
  },

  get bookmarks() {
    let bar = new AtulBar({
      id: "PersonalToolbar",
      type: TOOLBAR_BOOKMARKS,
      toolbar: this
    });
    let _bookmarks = {
      append: function(options) {
        bar.append(options);
      },
      remove: function(buttonID) {
        bar.remove(buttonID);
      }
    };
    this._unloaders.push(function() bar.unload());

    this.__defineGetter__("bookmarks", function () _bookmarks);
    return this.bookmarks;
  }
};


function AtulBar(options) {
  this._id = options.id;
  this._toolbar = options.toolbar;
  this._type = options.type;
  this._buttons = {};
  this._browserWatchers = {};
  this._windowCounts = {};

  // Unload each button in the toolbar. This add an |this.unload()| method.
  Extension.addUnloadMethod(this, function() {
    for (let id in this._browserWatchers) {
      this._browserWatchers[id].unload();
      delete this._browserWatchers[id];
    }
  });

}
AtulBar.prototype = {
  _id: null,
  _toolbar: null,
  _type: null,
  _buttons: null,
  _browserWatchers: null,
  _windowCounts: null,

  // Public API ///////////////////////////////////////////////////////////////

  append: function AtulBar_append(options) {
    // _checkIsValidButton throws an error if not given valid options
    this._checkIsValidButton(options);
    let self = this;
    // Track number of windows per button, not in each toolbar.
    this._windowCounts[options.id] = 0;
    // Add the browser watcher
    this._browserWatchers[options.id] = new BrowserWatcher({
      onLoad: function(window) {
        self._addButtonToWindow(window, options);
        self._windowCounts[options.id]++;
      },
      onUnload: function(window) {
        self._removeButtonFromWindow(window, self._generateChromeButtonID(options.id));
        // Explicitly check window count so we don't remove the id prematurely
        if (--self._windowCounts[options.id] == 0)
          delete self._buttons[options.id];
      }
    });
    this._buttons[options.id] = options;
  },

  remove: function AtulBar_remove(id) {
    if (!id in this._buttons)
      throw ERRORS.BUTTON_NOT_FOUND;

    this._browserWatchers[id].unload();
    delete this._browserWatchers[id];
  },

  // Helper Functions /////////////////////////////////////////////////////////

  _checkIsValidButton: function AtulBar__isValidButton(options) {
    // Errors are thrown as found, so only 1 error will be thrown at a time.

    // An id is required to identify the button within this toolbar
    if (!options.id)
      throw new Error(ERRORS.BUTTON_OPTIONS_ID_REQUIRED);

    // The button id cannot already be present within this toolbar
    if (options.id in this._buttons)
      throw new Error(ERRORS.BUTTON_OPTIONS_UNIQUE_ID);

    // Normal toolbars require an onclick function to call
    if (this._type == TOOLBAR_NORMAL && !options.onclick)
      throw new Error(ERRORS.BUTTON_OPTIONS_ONCLICK_REQUIRED);

    // Bookmark toolbars require a url to open
    if (this._type == TOOLBAR_BOOKMARKS && !options.url)
      throw new Error(ERRORS.BUTTON_OPTIONS_URL_REQUIRED);
  },

  _addButtonToWindow: function AtulBar__addButtonToWindow(window, options) {
    try {
      let button = window.document.createElement("toolbarbutton");
      button.setAttribute('label', options.label);
      button.setAttribute("tooltiptext", options.tooltip);
      button.setAttribute("image", options.image);
      // Make sure the id is unique in the context
      let chromeButtonID = this._generateChromeButtonID(options.id);
      button.setAttribute('id', chromeButtonID);
      // Specialize onclick
      let self = this;
      if (this._type == TOOLBAR_NORMAL) {
        // Call the onClick handler in the feature's sandbox.
        button.onclick = function() {
          self._toolbar._featureContext.doUnsafeCallback(options.onclick);
        }
      } else {
        button.onclick = function(event) {
          window.openUILink(self._toolbar._featureContext.urlFactory.makeUrl(options.url),
                            event, false, false, false, null, null);
        };
        button.setAttribute("class", "bookmark-item");
      }
      let bar = this._getToolbar(window);
      bar.appendChild(button);
    } catch (e) {
      Components.utils.reportError(e);
      throw e;
    }
  },

  _removeButtonFromWindow: function AtulBar__removeButtonFromWindow(window, buttonID) {
    let button = window.document.querySelector("#" + buttonID);
    if (button)
      button.parentNode.removeChild(button);
  },

  _generateChromeButtonID: function AtulBar__generateChromeButtonID(id) {
    return ["jetpackbutton", this._type, this._toolbar._featureContext.id, id].join("-");
  },

  _getToolbar: function AtulBar__getToolbar(window) {
    //XXXzpao Need to add ability to create toolbar.
    //XXXzpao Consider creating a <toolbaritem> per jetpack so we could
    //        presumably move buttons together.
    return window.document.querySelector("#" + this._id);
  }
};
