/* -*- Mode: js2; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* vim:set ts=2 sw=2 sts=2 et: */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Places Query API code.
 *
 * The Initial Developer of the Original Code is Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  David Dahl <ddahl@mozilla.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const EXPORTED_SYMBOLS = ["PlacesQuery"];

const Ci = Components.interfaces;
const Cc = Components.classes;
const Cu = Components.utils;

Cu.import("resource://gre/modules/XPCOMUtils.jsm");

// XPCOMUtils.defineLazyServiceGetter doesn't exist on Firefox 3.5
if (!XPCOMUtils.defineLazyServiceGetter) {
  XPCOMUtils.defineLazyServiceGetter =
  function XPCU_defineLazyServiceGetter(aObject, aName, aContract, aInterfaceName) {
    aObject.__defineGetter__(aName, function() {
      delete aObject[aName];
      return aObject[aName] = Cc[aContract].getService(Ci[aInterfaceName]);
    });
  };
}

let NetUtil;
try {
  Cu.import("resource://gre/modules/NetUtil.jsm");
}
catch (ex) {
  // Make this compatible with Firefox 3.5, by using io service instead of the
  // new NetUtil
  XPCOMUtils.defineLazyServiceGetter(this, "io",
                                     "@mozilla.org/network/io-service;1",
                                     "nsIIOService");
  NetUtil = {
    newURI: function newURI(url) {
      return io.newURI(url, null, null);
    }
  };
}

XPCOMUtils.defineLazyServiceGetter(this, "hs",
                                   "@mozilla.org/browser/nav-history-service;1",
                                   "nsINavHistoryService");

XPCOMUtils.defineLazyServiceGetter(this, "bs",
                                   "@mozilla.org/browser/nav-bookmarks-service;1",
                                   "nsINavBookmarksService");

XPCOMUtils.defineLazyServiceGetter(this, "ts",
                                   "@mozilla.org/browser/tagging-service;1",
                                   "nsITaggingService");

function strictTypeOf(aObj) {
  if (isArray(aObj)) {
    return "array";
  }

  if (isDate(aObj)) {
    return "date";
  }

  let _type = typeof aObj;
  if (_type) {
    return _type;
  }
  else {
    throw new Error(ERRORS.UNDETERMINED_TYPE);
  }
}

function isDate(aObj) {
  return (aObj.constructor.name == "Date");
}

function isArray(aObj) {
  return (aObj.constructor.name == "Array");
}

function isNotAString(aObj) {
  return typeof aObj != 'string';
}

const ERRORS = {
    NO_QUERY_CONFIG: "No QueryConfig was provided to the PlacesQuery"
  , NO_QUERY_PHRASE: "No Query Phrase was provided to the PlacesQuery"
  , NOT_READY: "Cannot fetch, query is not ready"
  , SEARCH_TYPE_NOT_SUPPORTED: "Unsupported search type"
  , BOOKMARK_CREATE_FAILED: "Could not create bookmark, make sure you provide a URL"
  , BOOKMARK_DOES_NOT_EXIST: "Bookmark does not exist"
  , BOOKMARK_EXISTS: "Bookmark already exists"
  , TYPE_UNIDENTIFIED: "Could not identify the object's type"
  , BAD_ARGUMENT: "Please provide correct argument type"
  , UNDETERMINED_TYPE: "Cannot determine object's type"
  , UNIMPLEMENTED: "Unimplemented method"
  , INVALID_URL: "Invalid URL"
  , CREATED_BOOKMARK_NOT_FOUND: "Could not find newly-created bookmark"
  , TAGS_INPUT_ERROR: "Tags input is invalid. String or Array required"
};

function PlacesQuery(aQueryConf, aCallback) {

  if (!aQueryConf) {
    this.queryConf = { phrase: "__init__",
                       where: 'bookmarks',
                       callback: aCallback };
  }
  else {
    this.queryConf = { callback: aCallback };
    for (let prop in aQueryConf) {
      this.queryConf[prop] = aQueryConf[prop];
    }
  }
  // QueryConf.where should always be history or bookmarks.
  if (['history', 'bookmarks'].indexOf(this.queryConf.where) == -1){
    throw new Error(ERRORS.SEARCH_TYPE_NOT_SUPPORTED);
  }
  if (!this.queryConf.phrase){
    // Just in case aQueryConf.phrase is set to null or undefined
    throw new Error(ERRORS.NO_QUERY_PHRASE);
  }
  // Run process and return Query.
  return this.handOff();
}

PlacesQuery.prototype = {

  handOff: function PlacesQuery_handOff() {
    // Hand off the query object to the QueryExecutor.
    let QueryX = new QueryExecutor(this.queryConf);
    return QueryX;
  }
};

function QueryExecutor(aQueryConf, aCallback) {
  // executes the queries that are configured by the QueryAPI
  if (!aQueryConf || typeof aQueryConf !== 'object') {
    throw new Error(ERRORS.NO_QUERY_CONFIG);
  }

  this.queryConf = aQueryConf;

  if (aQueryConf.phrase != "__init__") {
    this.find(aQueryConf, aCallback);
  }
}

QueryExecutor.prototype = {

  reconfigure:  function QE_reconfigure(aQueryConf, aCallback) {
    for (let prop in aQueryConf) {
      this.queryConf[prop] = aQueryConf[prop];
    }
    if (aCallback) {
      this.queryConf.callback = aCallback;
    }

    let self = this;

    this.queryOptions = hs.getNewQueryOptions();
    this.queryOptions.resultType = Ci.nsINavHistoryQueryOptions.RESULTS_AS_URI;
    this.query = hs.getNewQuery();

    if (this.queryConf.where === 'bookmarks') {
      this.queryOptions.queryType = Ci.nsINavHistoryQueryOptions.QUERY_TYPE_BOOKMARKS;
    }
    else if (this.queryConf.where === 'history') {
      this.queryOptions.queryType = Ci.nsINavHistoryQueryOptions.QUERY_TYPE_HISTORY;
    }

    if (this.queryConf.phrase) {
      this.query.searchTerms = this.queryConf.phrase;
    }
    else {
      throw new Error(ERRORS.NO_QUERY_PHRASE);
    }
  },

  // main public API lives here
  find: function QueryExecutor_find(aQueryConf, aCallback) {
    this.reconfigure(aQueryConf, aCallback);
    let callback = this.queryConf.callback;

    var root = hs.executeQuery(this.query, this.queryOptions).root;
    root.containerOpen = true;
    let count = root.childCount;
    let results = new Array();

    for (let i = 0; i < count; i++) {
      let child = root.getChild(i);
      results.push(new ResultItem(child));
    }
    if (callback) {
      callback(results);
    }
  },

  get bookmark() { return new QE_bookmarkFactory(); }

};

function QE_bookmarkFactory() {
  // getter that returns object that creates bookmarks
}

QE_bookmarkFactory.prototype = {
  createOne: function QE_bookmark_createOne(aObj) {
    if (!("url" in aObj) || !aObj.url) {
      throw new Error(ERRORS.BAD_ARGUMENT);
    }
    let self = this;
    let bmUri;
    try {
      bmUri = NetUtil.newURI(aObj.url);
    }
    catch(ex) {
      throw new Error(ERRORS.INVALID_URL);
    }

    // XXXddahl: check for aObj.folderName to find the correct folder to
    // place the new bookmarks into
    if (!("title" in aObj)) {
      aObj.title = "";
    }

    let tags = [];

    if ("tags" in aObj) {
      // if (["array", "string"].indexOf(strictTypeOf(aObj.tags) != -1)) {
      if (isArray(aObj.tags) || typeof aObj.tags == "string") {
        tags = tags.concat(aObj.tags);
      }
      else {
        throw new Error(ERRORS.TAGS_INPUT_ERROR);
      }
    }

    let id = bs.insertBookmark(bs.unfiledBookmarksFolder,
                               bmUri,
                               bs.DEFAULT_INDEX,
                               aObj.title);

    if (tags.length){
      ts.tagURI(bmUri, tags);
    }

    this.createdBookmarkIds.push(id);

    // look up the created bookmark, add it to the createdBookmarks array
    // use separate internal queryAPI object here to get the results
    let callback = function (results) {
      if (results.length < 1) {
        throw new Error(ERRORS.CREATED_BOOKMARK_NOT_FOUND);
      }
      let len = results.length;
      for (let i = 0; i < len; i++) {
        // filter out older bookmarks sharing the same uri.spec
        if (self.createdBookmarkIds.indexOf(results[i].id) > -1) {
          self.createdBookmarks.push(results[i]);
        }
      }
    };
    let queryConf = { where: "bookmarks",
                      phrase: bmUri.spec };
    let qapi = new PlacesQuery(queryConf, callback);

  },

  create: function QE_bookmark_create(aObj, aCallback) {
    // handle an array of potential bookmark objs or single objects
    // create the accumulator to collect the bookmarks created

    this.createdBookmarks = [];
    this.createdBookmarkIds = [];

    if (isArray(aObj)) {
      let len = aObj.length;
      for (let i = 0; i < len; i++) {
        this.createOne(aObj[i]);
      }
    }
    else {
      if (strictTypeOf(aObj) != 'object') {
        throw new Error(ERRORS.BAD_ARGUMENT);
      }
      this.createOne(aObj);
    }
    if (aCallback) {
      aCallback(this.createdBookmarks);
    }
  },

  exists: function QE_bookmark_exists(aUrl) {
    // returns true if url is bookmarked
    let _uri = NetUtil.newURI(aUrl);
    return bs.isBookmarked(_uri);
  },

  update: function QE_bookmark_update(aBookmark, aCallback) {
    // update the attributes of a bookmark that is already in the database
    throw new Error(ERRORS.UNIMPLEMENTED);
  }
};

const RESULT_ITEM_TYPE_OBJ = {
  0: 'uri',
  1: 'visit',
  2: 'full_visit',
  4: 'container',
  5: 'query',
  6: 'folder',
  7: 'separator',
  9: 'shortcut',
};

// ResultItem is a Places query node

function ResultItem(child) {
  // XXXddahl: QueryInterface for history or bookmark child!
  // create a resultItem from the child node
  this._rawNode = child;
}

const TAGS_SEPARATOR = ", ";

ResultItem.prototype = {
  // will have to refactor this to take into account the closed container
  children: function ResultItem_children() {
    if (this._rawNode.hasChildren) {
      let children = [];
      for (let i = 0; i < this._rawNode.childCount; i++) {
        children.push(new ResultItem(this._rawNode.getChild(i)));
      }
      return children;
    }
    else {
      return null;
    }
  },

  get tags() {
    try {
      return this._rawNode.tags.split(TAGS_SEPARATOR);
    }
    catch (ex) {
      return null;
    }
  },

  get genericType() {
    if (this._rawNode.lastModified > 0) {
      return 'bookmark';
    }
    else {
      return 'history';
    }
  },

  get siblingIndex() {
    try {
      return this._rawNode.bookmarkIndex;
    }
    catch(ex) {
      return -1;
    }

  },

  get id() {
    return this._rawNode.itemId;
  },

  get title() {
    return this._rawNode.title;
  },

  get uri() {
    return this._rawNode.uri;
  },

  get lastAccessTime() {
    if (this._rawNode.time) {
      return new Date((this._rawNode.time / 1000));
    }
    else {
      return null;
    }
  },

  get creationTime() {
    return new Date((this._rawNode.dateAdded / 1000));
  },

  get modificationTime() {
    try {
      return new Date((this._rawNode.lastModified / 1000));
    }
    catch(ex) {
      // must be a history item, just return the creation time
      return this.creationTime;
    }
  },

  get isBookmark() {
    return (this.genericType == "bookmark");
  },

  get nodeType() {
    let _type = this._rawNode.nodeType;
    return RESULT_ITEM_TYPE_OBJ[_type];
  }
};
